package edu.berkeley.sbp.meta;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.chr.*;
import edu.berkeley.sbp.misc.*;
import edu.berkeley.sbp.bind.*;
import java.util.*;
import java.lang.annotation.*;
import java.lang.reflect.*;
import java.io.*;

/** The java classes typically used to represent a parsed grammar AST; each inner class is a type of AST node. */
public class MetaGrammarBindings {

    /** A grammar (a set of nonterminals) */
    public static class GrammarNode extends HashMap<String,NonTerminalNode> {
        public @bind.as("Grammar") GrammarNode(NonTerminalNode[] nonterminals) {
            for(NonTerminalNode nt : nonterminals) this.put(nt.name, nt); }
        public String toString() {
            String ret = "[ ";
            for(NonTerminalNode nt : values()) ret += nt + ", ";
            return ret + " ]";
        }
        public Union build(String s, GrammarBindingResolver rm) {
            Context cx = new Context(this,rm);
            Union u = null;
            for(MetaGrammarBindings.NonTerminalNode nt : values()) {
                Union el = (Union)cx.get(nt.name);
                StringBuffer st = new StringBuffer();
                el.toString(st);
                if (nt.name.equals(s)) u = el;
            }
            return u;
        }
    }

    public abstract static class UnionNode extends ElementNode {
        public Seq[][] sequences;
        public void build(Context cx, Union u) {
            HashSet<Sequence> bad2 = new HashSet<Sequence>();
            for(int i=0; i<sequences.length; i++) {
                Seq[] group = sequences[i];
                Union u2 = new Union();
                if (sequences.length==1) u2 = u;
                for(int j=0; j<group.length; j++) {
                    group[j].build(cx, u2, false);
                }
                if (sequences.length==1) break;
                Sequence seq = Sequence.singleton(u2);
                for(Sequence s : bad2) {
                    s.lame = true;
                    seq = seq.not(s);
                }
                u.add(seq);
                bad2.add(Sequence.singleton(u2));
            }
        }
    }

    public static class NonTerminalNode extends UnionNode {
        public boolean rep;
        public String  name = null;
        public String sep = null;
        public @bind.as("NonTerminal") NonTerminalNode(@bind.arg String name, @bind.arg Seq[][] sequences) { this(name, sequences, false); }
        public NonTerminalNode(String name, Seq[][] sequences, boolean rep) { this(name, sequences, rep, null); }
        public NonTerminalNode(String name, Seq[][] sequences, boolean rep, String sep) {
            this.name = name;
            this.sequences = sequences;
            this.rep = rep;
            this.sep = sep;
        }
        public Element build(Context cx) { return cx.get(name); }
        public void build(Context cx, Union u) {
            if (!rep) { super.build(cx, u); return; }
            HashSet<Sequence> bad2 = new HashSet<Sequence>();

            Union urep = new Union();
            urep.add(Sequence.empty);
            urep.add(Sequence.singleton(new Element[] { cx.get(sep), u }, 1));

            for(int i=0; i<sequences.length; i++) {
                Seq[] group = sequences[i];
                Union u2 = new Union();
                if (sequences.length==1) u2 = u;
                for(int j=0; j<group.length; j++) {
                    Union u3 = new Union();
                    group[j].build(cx, u3, false);
                    Sequence s = Sequence.unwrap(new Element[] { u3, urep },
                                                 cx.rm.repeatTag(),
                                                 new boolean[] { false, false });
                    u2.add(s);
                }
                if (sequences.length==1) break;
                Sequence seq = Sequence.singleton(u2);
                for(Sequence s : bad2) {
                    s.lame = true;
                    seq = seq.not(s);
                }
                u.add(seq);
                bad2.add(Sequence.singleton(u2));
            }
        }
    }
    public static @bind.as("=") NonTerminalNode go(@bind.arg String name, @bind.arg Seq[][] sequences) { return new NonTerminalNode(name, sequences, true); }
    public static @bind.as("=") NonTerminalNode go(@bind.arg String name, @bind.arg String sep, @bind.arg Seq[][] sequences) {
        return new NonTerminalNode(name, sequences, true, sep);
    }

    public static class AnonUnionNode extends UnionNode {
        public @bind.as("(") AnonUnionNode(Seq[][] sequences) {
            this.sequences = sequences;
        }
        public Element build(Context cx) {
            Union ret = new Union();
            build(cx, ret);
            return ret;
        }
    }

    public static class Range {
        public @bind Range(char only) { first = only; last = only; }
        public @bind Range(char first, char last) { this.first = first; this.last = last; }
        public char first;
        public char last;
    }

    public static abstract class ElementNode {
        public String getLabel() { return null; }
        public String getOwnerTag() { return null; }
        public boolean drop() { return false; }
        public abstract Element build(Context cx);
    }

    public static class Drop extends ElementNode {
        public ElementNode e;
        public Drop(ElementNode e) { this.e = e; }
        public String getLabel() { return null; }
        public boolean drop() { return true; }
        public String getOwnerTag() { return e.getOwnerTag(); }
        public Element build(Context cx) { return e.build(cx); }
    }

    public static class Label extends ElementNode {
        public String label;
        public ElementNode e;
        public Label(String label, ElementNode e) { this.e = e; this.label = label; }
        public String getLabel() { return label; }
        public String getOwnerTag() { return e.getOwnerTag(); }
        public Element build(Context cx) { return e.build(cx); }
    }

    public static /*abstract*/ class Seq {
        HashSet<Seq> and = new HashSet<Seq>();
        HashSet<Seq> not = new HashSet<Seq>();
        ElementNode[] elements;
        ElementNode follow;
        String tag = null;
        boolean lame;
        public void append(ElementNode e) {
            ElementNode[] elements = new ElementNode[this.elements.length+1];
            System.arraycopy(this.elements, 0, elements, 0, this.elements.length);
            this.elements = elements;
            elements[elements.length-1] = e;
        }
        public Seq(ElementNode e) { this(new ElementNode[] { e }); }
        public Seq(ElementNode[] elements) { this.elements = elements; }
        public Seq tag(String tag) { this.tag = tag; return this; }
        public Seq follow(ElementNode follow) { this.follow = follow; return this; }
        public Seq dup() {
            Seq ret = new Seq(elements);
            ret.and.addAll(and);
            ret.not.addAll(not);
            ret.follow = follow;
            ret.tag = tag;
            return ret;
        }
        public Seq and(Seq s) { and.add(s); s.lame = true; return this; }
        public Seq andnot(Seq s) { not.add(s); s.lame = true; return this; }
        public Seq separate(ElementNode sep) {
            ElementNode[] elements = new ElementNode[this.elements.length * 2 - 1];
            for(int i=0; i<this.elements.length; i++) {
                elements[i*2]   = this.elements[i];
                if (i<this.elements.length-1)
                    elements[i*2+1] = new Drop(sep);
            }
            this.elements = elements;
            return this;
        }
        public Sequence build(Context cx, Union u, boolean lame) {
            Sequence ret = build0(cx, lame || this.lame);
            for(Seq s : and) { Sequence dork = s.build(cx, u, true); ret = ret.and(dork); }
            for(Seq s : not) { Sequence dork = s.build(cx, u, true); ret = ret.not(dork); }
            u.add(ret);
            ret.lame = lame;
            return ret;
        }
        public Sequence build0(Context cx, boolean lame) {
            boolean dropAll = lame;
            if (tag!=null && "()".equals(tag)) dropAll = true;
            Object[] labels = new Object[elements.length];
            boolean[] drops = new boolean[elements.length];
            Element[] els = new Element[elements.length];
            for(int i=0; i<elements.length; i++) {
                labels[i] = elements[i].getLabel();
                drops[i]  = elements[i].drop();
                els[i] = elements[i].build(cx);
                if (elements[i].getOwnerTag() != null)
                    tag = elements[i].getOwnerTag();
            }
            Sequence ret = null;
            if (dropAll)     ret = Sequence.drop(els, false);
            else {
                ret = cx.rm.tryResolveTag(tag, cx.cnt, els, labels, drops);
                if (ret == null) {
                    int idx = -1;
                    for(int i=0; i<els.length; i++)
                        if (!drops[i])
                            if (idx==-1) idx = i;
                            else throw new Error("multiple non-dropped elements in sequence: " + Sequence.drop(els,false));
                    if (idx != -1) ret = Sequence.singleton(els, idx);
                    else           ret = Sequence.drop(els, false);
                }
            }
            if (this.follow != null)
                ret.follow = infer(this.follow.build(cx));
            ret.lame = this.lame;
            return ret;
        }
    }
    public static @bind.as("&")   Seq  and2(Seq s,        Seq a) { return s.and(a); }
    public static @bind.as("&~")  Seq  andnot2(Seq s,     Seq a) { return s.andnot(a); }
    public static @bind.as("->")  Seq  arrow(Seq s, ElementNode e)                { return s.follow(e); }
    public static @bind.as("::")  Seq  tag(String tagname, Seq s)        { return s.tag(tagname); }
    public static @bind.as("/")   Seq  slash(Seq s, ElementNode e)                { return s.separate(e); }

    public static Seq  seq(ElementNode[] elements)               { return new Seq(elements); }
    public static @bind.as("Elements")  Seq  seq2(ElementNode[] elements)               { return new Seq(elements); }
    public static @bind.as        Seq  psx(Seq s)                        { return s; }
    public static @bind.as(":")   ElementNode   colon(String s, ElementNode e)             { return new Label(s, e); }
    public static @bind.as(")")   void close(String foo)                 { throw new Error("not supported"); }
    public static @bind.as("()")  ElementNode   epsilon()                         { return new Constant(Union.epsilon); }

    public static @bind.as("NonTerminalReference") class NonTerminalReferenceNode extends ElementNode {
        public @bind.arg String nonTerminal;
        public Element build(Context cx) { return cx.get(nonTerminal); }
    }

    public static class Literal extends Constant {
        public @bind Literal(@bind.arg String string) { super(CharRange.string(string)); }
        public boolean drop() { return true; }
    }

    public static                     class CharClass            extends ElementNode {
        Range[] ranges;
        public @bind.as("[") CharClass(Range[] ranges) { this.ranges = ranges; }
        public Element build(Context cx) {
            edu.berkeley.sbp.util.Range.Set set = new edu.berkeley.sbp.util.Range.Set();
            for(Range r : ranges)
                set.add(r.first, r.last);
            return CharRange.set(set);
        }
    }

    public static @bind.as("{")           class XTree                 extends ElementNode {
        public @bind.arg Seq body;
        public Element build(Context cx) {
            Union u = new Union();
            Sequence s = body.build(cx, u, false);
            Union u2 = new Union();
            u2.add(Sequence.singleton(new Element[] {
                CharRange.leftBrace,
                cx.get("ws"),
                u,
                cx.get("ws"),
                CharRange.rightBrace
            }, 2));
            return u2;
        }
    }

    public static class Rep extends ElementNode {
        public ElementNode e, sep;
        public boolean zero, many, max;
        public Rep(ElementNode e, ElementNode sep, boolean zero, boolean many, boolean max) {
            this.e = e; this.sep = sep; this.zero = zero; this.many = many; this.max = max;}
        public Element build(Context cx) {
            return (!max)
                ? Sequence.repeat(e.build(cx),        zero, many, sep==null ? null : sep.build(cx), cx.rm.repeatTag())
                : sep==null
                ? Sequence.repeatMaximal(infer(e.build(cx)), zero, many,                                   cx.rm.repeatTag())
                : Sequence.repeatMaximal(e.build(cx),                    zero, many, infer(sep.build(cx)), cx.rm.repeatTag());
        }
    }
    public static class Constant extends ElementNode {
        Element constant;
        public Constant(Element constant) { this.constant = constant; }
        public Element build(Context cx) { return constant; }
    }
    public abstract static class PostProcess extends ElementNode {
        ElementNode e;
        public PostProcess(ElementNode e) { this.e = e; }
        public Element build(Context cx) { return postProcess(e.build(cx)); }
        public abstract Element postProcess(Element e);
    }

    // FIXME: it would be nice if we could hoist this into "Rep"
    public static @bind.as("++")  ElementNode plusmax(final ElementNode e)                     { return new Rep(e, null, false, true, true); }
    public static @bind.as("+")   ElementNode plus(final ElementNode e)                        { return new Rep(e, null, false, true, false); }
    public static @bind.as("++/") ElementNode plusmaxfollow(final ElementNode e, final ElementNode sep) { return new Rep(e, sep,  false, true, true); }
    public static @bind.as("+/")  ElementNode plusfollow(final ElementNode e, final ElementNode sep)    { return new Rep(e, sep,  false, true, false); }
    public static @bind.as("**")  ElementNode starmax(final ElementNode e)                     { return new Rep(e, null, true,  true, true); }
    public static @bind.as("*")   ElementNode star(final ElementNode e)                        { return new Rep(e, null, true,  true, false); }
    public static @bind.as("**/") ElementNode starmaxfollow(final ElementNode e, final ElementNode sep) { return new Rep(e, sep,  true,  true, true); }
    public static @bind.as("*/")  ElementNode starfollow(final ElementNode e, final ElementNode sep)    { return new Rep(e, sep,  true,  true, false); }
    public static @bind.as("?")   ElementNode question(final ElementNode e)                    { return new Rep(e, null, true,  true, false); }

    public static @bind.as("!")   ElementNode bang(final ElementNode e)                        { return new Drop(e); }

    public static @bind.as("^")   ElementNode caret(final String s) {
        return new Drop(new Constant(CharRange.string(s)) {
                public String getOwnerTag() { return s; }
            });
    }

    public static @bind.as("~")   ElementNode tilde(final ElementNode e) {
        return new PostProcess(e) {
                public Element postProcess(Element e) {
                    return infer((Topology<Character>)Atom.toAtom(e).complement()); 
                } }; }

    public static @bind.as("Word")        String word(String s) { return s; }
    public static @bind.as("Quoted")      String quoted(String s) { return s; }
    public static @bind.as("escaped")     String c(char c) { return c+""; }
    public static @bind.as("EmptyString") String emptystring() { return ""; }
    public static @bind.as("\n")          String retur() { return "\n"; }
    public static @bind.as("\r")          String lf() { return "\r"; }

    static Atom infer(Element e)  { return infer((Topology<Character>)Atom.toAtom(e)); }
    static Atom infer(Topology<Character> t) { return new CharRange(new CharTopology(t)); }

    public static class Context {
        public HashMap<String,Union> map = new HashMap<String,Union>();
        public GrammarNode grammar;
        public String cnt = null;
        public GrammarBindingResolver rm;
        public Context(GrammarNode g, GrammarBindingResolver rm) {
            this.grammar = g;
            this.rm = rm;
        }
        public Union build() {
            Union ret = null;
            for(NonTerminalNode nt : grammar.values()) {
                Union u = get(nt.name);
                if ("s".equals(nt.name))
                    ret = u;
            }
            return ret;
        }
        public Context(Tree t, GrammarBindingResolver rm) {
            this.rm = rm;
            Tree.TreeFunctor<Object,Object> red = (Tree.TreeFunctor<Object,Object>)t.head();
            this.grammar = (GrammarNode)red.invoke(t.children());
        }
        public Union peek(String name) { return map.get(name); }
        public void  put(String name, Union u) { map.put(name, u); }
        public Union get(String name) {
            Union ret = map.get(name);
            if (ret != null) return ret;
            ret = new Union(name);
            map.put(name, ret);
            NonTerminalNode nt = grammar.get(name);
            if (nt==null) {
                System.err.println("*** warning could not find " + name);
            } else {
                String old = cnt;
                cnt = name;
                nt.build(this, ret);
                cnt = old;
            }
            return ret;
        }

    }

}
